/*

Copyright (C) 2005, 2006 Christian Kreibich <christian@whoop.org>.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <regex.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>

#include <libnd.h>
#include <plugins/libnd_traceset.h>
#include <plugins/libnd_conntrack.h>
#include <protocols/tcp/libnd_tcp.h>

/* Interval in seconds between garbage collection of connection table */
#define LND_APPDEMUX_CONN_TABLE_GC_INTERVAL 3*60

typedef struct lnd_flow_data {
  GList         *packets;
  pcap_dumper_t *out_pcap;
} LND_FlowData;

static char *ad_prefix = "appdemux";
static char *ad_tethereal = NULL;
static pid_t ad_child;
static int   ad_fd;
static regex_t ad_re;

/* Hashtable for looking up output dumpers. */
static GHashTable *ad_out_pcaps; /* map<char*, pcap_dumper_t*> */

/* Connection state tracker. */
LND_ConnTable *ad_conns;

/* Last time we checked the connection table for dead connections. */
struct bpf_timeval ad_last_gc;

/* User flag: maximum size of connection table. Disabled by default. */
unsigned int ad_conn_table_maxsize;

static gboolean
ad_find_tethereal(void)
{
  char *path, *ptr, *next;

  if (ad_tethereal && libnd_misc_can_exec(ad_tethereal))
    return TRUE;

  path = ptr = strdup(getenv("PATH"));

  while ( (next = strchr(ptr, ':')))
    {
      char command[MAXPATHLEN];
      *next = '\0';

      snprintf(command, MAXPATHLEN, "%s/tethereal", ptr);
      command[MAXPATHLEN-1] = '\0';

      if (libnd_misc_can_exec(command))
	{
	  ad_tethereal = strdup(command);
	  free(path);
	  D(("Using tethereal at '%s'\n", ad_tethereal));
	  return TRUE;
	}

      ptr = next + 1;
    }
  
  free(path);
  return FALSE;
}


static void
ad_help(void)
{
  printf("This plugin demultiplexes traffic based on application-layer protocol content.\n"
	 "\n"
	 "USAGE: lndtool -r trace-info [--prefix|-p <PREFIX>] [--names-file|-f <FILE>]\n"
	 "                             [--tethereal|-t <FILE>] [--max-flows|-m <NUM>] [<TRACE2> <TRACE3> ...]\n"
	 "\n"
	 "  --help, -h, -?           This message.\n"
	 "  --prefix|-p <PREFIX>     Prefix to prepend to output files. Default: 'appdemux'.\n"
	 "  --names-file|-f <FILE>   Take filenames of traces from flatfile FILE.\n"
	 "  --tethereal|-t <FILE>    Use tethereal executable given in FILE.\n"
	 "  --max-flows|-m <NUM>     Maximum number of flows in connection table before evictions.\n"
	 "                           By default, only timeouts are used for expiration.\n");
}


static gboolean
ad_setup_tethereal(LND_Trace *trace)
{
  int fd[2];

  if (socketpair(AF_UNIX, SOCK_STREAM, 0, fd) < 0)
    {
      fprintf(stderr, "Could not create socket pair -- exiting.\n");
      return FALSE;
    }
  
  if ( (ad_child = fork()) < 0)
    {
      fprintf(stderr, "Fork error -- exiting.\n");
      return FALSE;
    }
  else if (ad_child > 0) /* Parent */
    {
      close(fd[1]);
      ad_fd = fd[0];

      D(("Child %i forked.\n", ad_child));

      if (fcntl(ad_fd, F_SETFL, O_NONBLOCK) < 0)
	{
	  fprintf(stderr, "Can not fcntl socket -- exiting.\n");
	  return FALSE;
	}

      if (libnd_misc_writen(ad_fd, (guchar *) &(trace->tcpdump.pfh),
			    sizeof(struct pcap_file_header))
	  != sizeof(struct pcap_file_header))
	{
	  fprintf(stderr, "Write error in pipe when sending header\n");
	  return FALSE;	  
	}
      
      D(("Trace header written.\n"));
      return TRUE;
    }
  else
    {
      /* The child replaces the process image with tethereal */
      close(fd[0]);
      ad_fd = fd[1];

      if (ad_fd != STDIN_FILENO)
	{
	  if (dup2(ad_fd, STDIN_FILENO) != STDIN_FILENO)
	    {
	      fprintf(stderr, "stdin pipe error\n");
	      exit(-1);
	    }
	}
      if (ad_fd != STDOUT_FILENO)
	{
      	  if (dup2(ad_fd, STDOUT_FILENO) != STDOUT_FILENO)
	    {
	      fprintf(stderr, "stdout pipe error\n");
	      exit(-1);
	    }
	}
      
      if (execlp(ad_tethereal, "tethereal", "-l", "-i", "-", NULL) < 0)
	{
	  fprintf(stderr, "tcpdump error launching %s: %s\n", ad_tethereal, strerror(errno));
	  exit(-1);
	}
    }

  return TRUE;  
}

static void
ad_shutdown_tethereal(void)
{
  if (ad_child <= 0)
    return;
  
  kill(ad_child, SIGKILL);
  close(ad_fd);

  if (waitpid(ad_child, NULL, 0) != ad_child)
    printf("Error in waitpid() for tethereal process.\n");

  ad_child = 0;
  ad_fd = 0;
}

static gboolean
ad_send_packet(const LND_Packet *packet)
{
  fd_set       fdset;
  
  FD_ZERO(&fdset);
  FD_SET(ad_fd, &fdset);
  
  if (select(ad_fd + 1, NULL, &fdset, NULL, NULL) <= 0)
    goto error_exit;
  
  if (write(ad_fd, &(packet->ph), sizeof(struct pcap_pkthdr)) != sizeof(struct pcap_pkthdr))
    goto error_exit;
  
  if (write(ad_fd, packet->data, packet->ph.caplen) != (int) packet->ph.caplen)
    goto error_exit;
  
  return TRUE;
  
 error_exit:
  printf("Write error in pipe: %s\n", strerror(errno));
  return FALSE;
}

static pcap_dumper_t*
ad_lookup_dumper(LND_Trace *trace, const char *name)
{
  pcap_dumper_t* out_pcap = NULL;
      
  if (! (out_pcap = g_hash_table_lookup(ad_out_pcaps, name)))
    {
      char out_file[MAXPATHLEN];
      
      snprintf(out_file, MAXPATHLEN, "%s-%s.trace", ad_prefix, name);
      out_file[MAXPATHLEN-1] = '\0';

      /* FIXME -- use libnd_trace_get_pcap_handle() in future releases. */
      out_pcap = pcap_dump_open(pcapnav_pcap(trace->tpm->base->pcn), out_file);
      g_hash_table_insert(ad_out_pcaps, strdup(name), out_pcap);
      D(("Created output dumper '%s'\n", out_file));
    }

  return out_pcap;
}

static LND_FlowData*
ad_flowdata_new(void)
{
  LND_FlowData *data;

  if (! (data = g_new0(LND_FlowData, 1)))
    return NULL;

  return data;
}

static void
ad_flowdata_free(LND_FlowData* data)
{
  GList *l;

  if (! data)
    return;

  for (l = data->packets; l; l = g_list_next(l))
    {
      LND_Packet *packet = l->data;
      libnd_packet_free(packet);
    }

  g_list_free(data->packets);
  free(data);
}

/* Garbage-collect the connection table.
 */
static gboolean
ad_conn_table_gc_cb(LND_ConnID *conn, LND_Packet *current)
{
  if (libnd_conn_is_dead(conn, current))
    {
      /* There is a possibilty of more than one connection in 
       * the table matching this connection -- to prevent segfaults
       * make sure we only nuke the one we mean.
       */
      if (libnd_conn_table_remove(ad_conns, conn) == conn)
	{
	  LND_FlowData *fd = (LND_FlowData*) libnd_conn_data_remove(conn, "fd");
	  ad_flowdata_free(fd);
	  libnd_conn_free(conn);
	}

      return TRUE; /* Keep going. */
    }
  
  /* If the connection isn't dead yet, we abort the callback iteration,
   * because all future flows will be younger than the current one:
   * we're iterating backward through the age list.
   */
  return FALSE;
}

static void
ad_conn_table_gc(LND_Packet *current)
{
  libnd_conn_table_foreach_oldest(ad_conns, (LND_ConnFunc) ad_conn_table_gc_cb, current);
    
  /* If the connection table is still too big, evict oldest connections
   * until we're okay again.
   */
  if (ad_conn_table_maxsize > 0)
    {
      while (libnd_conn_table_size(ad_conns) > ad_conn_table_maxsize)
	{
	  LND_ConnID *conn = libnd_conn_table_get_oldest(ad_conns);
	  
	  if (libnd_conn_table_remove(ad_conns, conn) == conn)
	    {
	      LND_FlowData *fd = (LND_FlowData*) libnd_conn_data_remove(conn, "fd");
	      ad_flowdata_free(fd);
	      libnd_conn_free(conn);
	    }
	}
    }
}


/* The main routine for a single trace -- iterates over all packets
 * in the given trace, maintains state for all flows in the trace,
 * and dumps flows into pcap files depending on the app-layer protocol
 * that tethereal has determined for the flow.
 */
static gboolean
ad_demux(LND_TraceSet *set, LND_Trace *trace, void *unused)
{
  LND_PacketIterator  pit;
  LND_Packet *packet;
  LND_ConnID *conn;
  char line[MAXPATHLEN];
  char proto[128];
  regmatch_t rm[2];
  int len;
  pcap_dumper_t *out_pcap = NULL;
  LND_FlowData *fd = NULL;
  struct bpf_timeval tv;

  line[0] = '\0';

  if (! ad_setup_tethereal(trace))
    {
      printf("Couldn't run tethereal on input trace '%s'.\n"
	     "Try specifying a tethereal executable manually using --tethereal|-t.\n",
	     trace->filename);
      return TRUE;	     
    }

  /* Iterate over entire individual trace, read-only. */
  for (libnd_pit_init_mode(&pit, trace, LND_PACKET_IT_AREA_R);
       libnd_pit_get(&pit); libnd_pit_next(&pit))
    {
      packet = libnd_pit_get(&pit);

      /* Make sure we have connection state for this packet's flow */
      if (! (conn = libnd_conn_table_lookup(ad_conns, packet)))
	{
	  if (! (conn = libnd_conn_new(packet)))
	    {
	      /* It's a non-IP packet -- we dump all those in a
	       * separate output trace.
	       */
	      out_pcap = ad_lookup_dumper(trace, "others");
	      pcap_dump((u_char*) out_pcap, &(packet->ph), packet->data);
	      continue;
	    }

	  libnd_conn_table_add(ad_conns, conn);

	  /* Create flow-specific state and associate it with the
	   * connection ID.
	   */
	  fd = ad_flowdata_new();
	  libnd_conn_data_set(conn, "fd", fd);
	}
      
      /* We now definitely have connection state with associated flow data */

      libnd_conn_update(conn, packet);
      fd = (LND_FlowData*) libnd_conn_data_get(conn, "fd");

      if (! fd->out_pcap)
	{
	  /* Send packet to tethereal, obtain tethereal output. */
	  ad_send_packet(packet);
	  len = libnd_misc_readline(ad_fd, line, MAXPATHLEN);
	  line[len] = '\0'; /* terminate -- should be done by libnd_misc_readline(). */
	  
	  /* Parse out the app-layer protocol name */
	  if (regexec(&ad_re, line, 2, rm, 0) != 0)
	    {
	      D(("Regex failed on '%s'\n", line));
	      continue;
	    }
	  
	  memcpy(proto, line + rm[1].rm_so, rm[1].rm_eo - rm[1].rm_so);
	  proto[rm[1].rm_eo - rm[1].rm_so] = '\0';
	  
	  /* Ensure output dumper for this protocol */
	  out_pcap = ad_lookup_dumper(trace, proto);

	  /* If it's TCP, we treat the connection specially because we want to wait
	   * until app-layer data has been exchanged so we know that tethereal's had
	   * a chance to determine the protocol.
	   */
	  if (conn->proto == IPPROTO_TCP)
	    {
	      struct ip* iphdr;
	      struct tcphdr *tcphdr;

	      /* Connection is established. Once we have actual payload, we
	       * assume tethereal is able to figure out the payload. If not,
	       * we throw the connection's packets on the "TCP" pile.
	       */
	      if (libnd_tcp_get_headers(packet, &iphdr, &tcphdr) &&
		  libnd_tcp_get_payload_length(iphdr, tcphdr) > 0)
		{
		  fd->out_pcap = out_pcap;
		}
	      else
		{
		  fd->packets = g_list_append(fd->packets, libnd_packet_duplicate(packet));
		  continue;
		}
	    }
	  else
	    {
	      /* For all other protocols we assume tethereal is able to
	       * identify them with the first packet.
	       */
	      fd->out_pcap = out_pcap;
	    }
	}
      /* If we have an output dumper, just write out the packet
       * and skip to the next one.
       */
      if (fd->out_pcap)
	{
	  if (fd->packets)
	    {
	      GList *l;
	      
	      /* Dump any buffered packets and delete them. */
	      for (l = fd->packets; l; l = g_list_next(l))
		{
		  LND_Packet *copy = l->data;
		  pcap_dump((u_char*) fd->out_pcap, &(copy->ph), copy->data);
		  libnd_packet_free(copy);	      
		}

	      g_list_free(fd->packets);
	      fd->packets = NULL;
	    }
	  
	  /* Dump the current packet. */
	  pcap_dump((u_char*) fd->out_pcap, &(packet->ph), packet->data);
	}

      /* If the connection is dead, remove it right away and clean up
       * in order to keep the connection table as small as possible.
       */
      if (libnd_conn_is_dead(conn, packet))
	{
	  libnd_conn_table_remove(ad_conns, conn);
	  libnd_conn_free(conn);
	  ad_flowdata_free(fd);
	}

      /* If we haven't garbage-collected the connection table in
       * LND_APPDEMUX_CONN_TABLE_GC_INTERVAL seconds, do so now.
       */
      pcapnav_timeval_sub(&packet->ph.ts, &ad_last_gc, &tv);
      if (tv.tv_sec >= LND_APPDEMUX_CONN_TABLE_GC_INTERVAL)
	{
	  ad_conn_table_gc(packet);
	  ad_last_gc.tv_sec = packet->ph.ts.tv_sec;
	  ad_last_gc.tv_usec = packet->ph.ts.tv_usec;
	}
    }
  
  ad_shutdown_tethereal();

  return TRUE;
  TOUCH(set); TOUCH(unused);
}

static gboolean
ad_out_pcaps_cleanup_cb(gpointer key, gpointer value, gpointer user_data)
{
  g_free(key); /* char * */
  pcap_dump_close((pcap_dumper_t *) value);

  return TRUE;
  TOUCH(user_data);
}

/* ---- Plugin symbols below ------------------------------------------- */

const char *
name(void)
{
  return "AppDemux";
}

const char *
author(void)
{
  return "Christian Kreibich, <christian@whoop.org>";
}

const char *
version(void)
{
  return "0.2";
}


gboolean
init(void)
{
  /* Regex for "tethereal -i -" output (no frame numbering). Example:
   * '  0.211530 192.168.100.28 -> 200.33.146.222 DNS Standard query A nsmex2.uninet.net.mx'
   *
   *                   <-- time --->      <- src IP -->      ->      <- dst IP -->      < proto > */
  char *regex = "[\t ]*[[:digit:].]+[\t ]+[[:digit:].]+[\t ]+->[\t ]+[[:digit:].]+[\t ]+([^\t ]+) ";

  if (regcomp(&ad_re, regex, REG_EXTENDED) != 0)
    {
      D(("Regex compilation failed: %s\n", strerror(errno)));
      return FALSE;
    }

  ad_out_pcaps = g_hash_table_new(g_str_hash, g_str_equal);
  ad_conns = libnd_conn_table_new(LND_CONN_TABLE_INCLUDE_DEAD);

  if (! libnd_plugin_find("Trace-Set"))
    {
      D(("Trace-Set plugin not found\n"));
      return FALSE;
    }

  if (! libnd_plugin_find("Conntrack"))
    {
      D(("Conntrack plugin not found\n"));
      return FALSE;
    }

  if (! ad_find_tethereal())
    {
      printf("Couldn't find usable tethereal executable. Make sure\n"
	     "it is installed and found in your PATH, or try --tethereal.\n");
      return FALSE;
    }

  D(("AppDemux initialization complete.\n"));
  
  return TRUE;
}

gboolean
run(LND_Trace *trace, LND_PluginArgs *args)
{
  LND_TraceSet *traces;
  int i, files = 0;

  if (! (traces = libnd_traceset_new()))
    return FALSE;

  for (i = 0; i < args->argc; i++)
    {
      if (!strcmp(args->argv[i], "-h")     ||
	  !strcmp(args->argv[i], "--help") ||
	  !strcmp(args->argv[i], "-?"))
	{
	  ad_help();
	  return FALSE;
	}
      else if (!strcmp(args->argv[i], "-p") ||
	       !strcmp(args->argv[i], "--prefix"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass a string to this option.\n");
	      return FALSE;
	    }
	  
	  ad_prefix = args->argv[i];
	}
      else if (!strcmp(args->argv[i], "-f") ||
	       !strcmp(args->argv[i], "--names-file"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass the name of a file that contains\n"
		     "a list of trace files, one filename per line, when\n"
		     "using the --names-file|-f option.\n");
	      return FALSE;
	    }
	  
	  if (! libnd_misc_can_read(args->argv[i]))
	    printf("Could not open '%s'\n", args->argv[i]);
	  else if (! libnd_traceset_add_trace_name_list(traces, args->argv[i]))
	    printf("Could not add all traces provided in '%s'\n", args->argv[i]);
	  else
	    files++;
	}
      else if (!strcmp(args->argv[i], "-t") ||
	       !strcmp(args->argv[i], "--tethereal"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass a string to this option.\n");
	      return FALSE;
	    }
	  
	  ad_tethereal = args->argv[i];
	}
      else if (!strcmp(args->argv[i], "-m") ||
	       !strcmp(args->argv[i], "--maxflows"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass an integer to this option.\n");
	      return FALSE;
	    }
	  
	  ad_conn_table_maxsize = atoi(args->argv[i]);
	}
      else if (args->argv[i][0] != '-')
	{
	  if (! libnd_traceset_add_trace_name(traces, args->argv[i]))
	    printf("Trace file %s not added.\n", args->argv[i]);
	  else
	    files++;
	}
    }
  
  if (args->argc == 0)
    {
      ad_help();
      return FALSE;
    }

  libnd_traceset_foreach(traces, ad_demux, NULL);  
  libnd_traceset_free(traces);

  /* Close all output traces properly. */
  g_hash_table_foreach_remove(ad_out_pcaps, ad_out_pcaps_cleanup_cb, NULL);

  return TRUE;
  TOUCH(trace);
}
